#-----------------------------------------------------------------------------
# Group to Morph v.2
# Feb. 13, 2004
# Copyright (c) 2004 by David G. Drumright (ockham)

#-----------------------------------------------------------------------------

import poser
import string
import math
from Tkinter import *
scene = poser.Scene()

#-----------------------------------------------------------------------------
# Global vars
#-----------------------------------------------------------------------------

global OneActor
global OneGeom
global SelectedGroups
global DeltaVerts
global SymmVertIndexes
global RawFacets
global RawVerts
global SortedVerts
global Smoothable
global CenterX
global CenterY
global CenterZ

OneActor=0
OneGeom=0
Smoothable=0  # When using Ridge, smoothing is unavailable.
RawFacets={}
RawVerts=[]  # This is a list, not a dict.  Nec for Ridge, to get dups
SelectedGroups={} # index is number in list, content is name.
SortedVerts={}
DeltaVerts={} # Runs parallel to SortedVerts
SymmVertIndexes={}

# These are set up as dicts so we can
# make non-contiguous arrays.

CenterX=0.0
CenterY=0.0
CenterZ=0.0
CoarseInc = 0.001
FineInc = 0.0001

#-----------------------------------------------------------------------------

def InitFirst():
    # Check the basics before putting anything up.
    global OneActor
    global OneGeom

    OneActor = scene.CurrentActor()  # There's always a CurrentActor
    try:
        # Get the Geom of the chosen actor:
        OneGeom = OneActor.Geometry()
    except:
        raise "Chosen part doesn't have a mesh."

    # Get the group list so we can let you select.
    # You may want to use one of the pre-existing
    # groups, or you may have made your own.

    if OneGeom.Groups() == None:
        raise "No groups in this actor"

#-----------------------------------------------------------------------------

def LocateAllVerts():
    # This part of the init is called AFTER you select some groups.
    global RawFacets
    global RawVerts

    # Start the Facets and Verts lists with just the basics,
    # and the default markings for Group and Layer.
    # Reminder: use only the group-list index, not the group name,
    # in places where group needs to be an index.
    group = -1    # default index: real one will be >=0
    dist  = -1.0  # default net distance: real will be >=0.0
    X = 0.0
    Y = 0.0
    Z = 0.0
    # Fill in all with standard values, so that the ungrouped facets
    # and verts will be left distinct after finding the grouped ones.
    for F in range(OneGeom.NumPolygons()):
        RawFacets[F]=(F,group)

    # Go through the Polygons (facets) of the Geom, by Groups.
    # Reminder: this will give dups of V index.  This is what we
    # want at this stage so we can spot Ridges.  When we sort for
    # non-Ridge we eliminate the dups.
    SetList = OneGeom.Sets()
    for GroupIndex in SelectedGroups.keys():
        GroupName = SelectedGroups[GroupIndex] # optimize
        for FacetIndex in range(OneGeom.NumPolygons()):
            OneFacet = OneGeom.Polygon(FacetIndex)
            if OneFacet.InGroup(GroupName):
                F,dummyG = RawFacets[FacetIndex]
                RawFacets[FacetIndex] = F,GroupIndex  # Add the group info
                First = OneFacet.Start()
                Last = First + OneFacet.NumVertices()
                for V in SetList[First:Last]: # This is a list of vertex indexes in the poly
                    # Get the values:
                    OneVert = OneGeom.Vertex(V)
                    X = OneVert.X()
                    Y = OneVert.Y()
                    Z = OneVert.Z()
                    # Fill in the Group and XYZ info.
                    RawVerts.append((V,GroupIndex,dist,X,Y,Z))
                    # Note that this can include the same V more than once,
                    # if that V was in more than one group.

#-----------------------------------------------------------------------------

def SortForJoin():
    global SortedVerts
    global DeltaVerts
    global CenterX
    global CenterY
    global CenterZ
    # For the Joined set, simply pare down the RawVerts,
    # to the in-group members.
    SumX=0.0
    SumY=0.0
    SumZ=0.0
    Count=0.0
    for i in range(len(RawVerts)):
         V,G,D,X,Y,Z = RawVerts[i]
         if G >= 0:  # Leave out the ungrouped
             SortedVerts[V] = V,G,D,X,Y,Z
             DeltaVerts[V] = V,G,D,0.0,0.0,0.0
             # Form group center for use in Contract/Expand
             SumX = SumX + X
             SumY = SumY + Y
             SumZ = SumZ + Z
             Count=Count+1.0
    if Count > 0.0:
        CenterX=SumX/Count
        CenterY=SumY/Count
        CenterZ=SumZ/Count
    else:
        CenterX=0.0
        CenterY=0.0
        CenterZ=0.0

#-----------------------------------------------------------------------------

def SortForRidge():
    global SortedVerts
    global DeltaVerts
    SortByV = [] # temp list
    AOVSub = []  # temp list
    for (V,G,D,X,Y,Z) in RawVerts:
        if G >= 0:  # Eliminate the outofgroups
            SortByV.append(V,G) # Just pick up the V and G
    SortByV.sort() # This sorts by V, first member of tuple

    # Now that things are in order, we can just look
    # for members with same V and different [selected] Gs.  Any time
    # we find a pair with V1==V2 and G1 not same as G2, we put
    # the first of the pair (arbitrarily, doesn't really matter)
    # into the final list of Ridgeverts.

    for i in range(len(SortByV)-1): # Stop short of end
        V1,G1 = SortByV[i]
        V2,G2 = SortByV[i+1]
        if (V1==V2) and (G1 != G2): #Same vert number in two groups
            AOVSub.append(V1)

    # Use this list of V's to set up the Deltas.  This condenses
    # the RawVerts list down to a dict containing only the
    # verts shared by two groups.
    for i in range(len(RawVerts)):
        (V,G,D,X,Y,Z) = RawVerts[i]
        if V in AOVSub: # Choose only the verts listed as edgy
            SortedVerts[V] = V,G,D,X,Y,Z
            DeltaVerts[V]  = V,G,D,0.0,0.0,0.0

#-----------------------------------------------------------------------------
# The Displace routines actually move the verts so you can see what's
# going on.  They also change all verts in the ChangedVerts array.
# Only the orig side is moved; the Symm doesn't do anything until
# the Spawn actions.
#-----------------------------------------------------------------------------

def DisplaceX(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        OneVert.SetX(OneVert.X()+delta)
        DeltaVerts[V] = (dummyV,G,D,delta,OrigY,OrigZ)

#-----------------------------------------------------------------------------

def DisplaceY(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        OneVert.SetY(OneVert.Y()+delta)
        DeltaVerts[V] = (dummyV,G,D,OrigX,delta,OrigZ)

#-----------------------------------------------------------------------------

def DisplaceZ(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        dummyV,G,D,OrigX,OrigY,OrigZ = DeltaVerts[V]
        OneVert.SetZ(OneVert.Z()+delta)
        DeltaVerts[V] = dummyV,G,D,OrigX,OrigY,delta

#-----------------------------------------------------------------------------

def DisplaceBulge(delta):
    global DeltaVerts
    Norms = OneGeom.Normals()
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        dX= Norms[V].X() * delta
        dY= Norms[V].Y() * delta
        dZ= Norms[V].Z() * delta
        OneVert.SetX(OneVert.X()+dX)
        OneVert.SetY(OneVert.Y()+dY)
        OneVert.SetZ(OneVert.Z()+dZ)
        DeltaVerts[V] = (dummyV,G,D,dX,dY,dZ)

#-----------------------------------------------------------------------------

def DisplaceXscale(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        # Find direction of this vert from Group center.
        # If this vert is farther from Group center, move it farther
        # in the same direction.
        dX=(OneVert.X()-CenterX ) * delta * 5.0
        OneVert.SetX(OneVert.X()+dX)
        DeltaVerts[V] = (dummyV,G,D,dX,OrigY,OrigZ)

#-----------------------------------------------------------------------------

def DisplaceYscale(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        # Find direction of this vert from Group center.
        # If this vert is farther from Group center, move it farther
        # in the same direction.
        dY=(OneVert.Y()-CenterY ) * delta * 5.0
        OneVert.SetY(OneVert.Y()+dY)
        DeltaVerts[V] = (dummyV,G,D,OrigX,dY,OrigZ)

#-----------------------------------------------------------------------------

def DisplaceZscale(delta):
    global DeltaVerts
    for V in DeltaVerts.keys():
        OneVert = OneGeom.Vertex(V)
        (dummyV,G,D,OrigX,OrigY,OrigZ) = DeltaVerts[V]
        # Find direction of this vert from Group center.
        # If this vert is farther from Group center, move it farther
        # in the same direction.
        dZ=(OneVert.Z()-CenterZ ) * delta * 5.0
        OneVert.SetZ(OneVert.Z()+dZ)
        DeltaVerts[V] = (dummyV,G,D,OrigX,OrigY,dZ)

#-----------------------------------------------------------------------------

def SpawnResult(Name,Symm):
    global DeltaVerts
    global SymmVertIndexes
    global OneGeom
    DeltaX=0.0
    DeltaY=0.0
    DeltaZ=0.0

    ArraySize = OneGeom.NumVertices()
    NumDeltas = len(DeltaVerts)

    # Set all existing MTs to zero before starting.  Otherwise they will be
    # mysteriously affected by the new morph.  Reminder: don't do this
    # earlier; sometimes you need to have other morphs active so you
    # can see what's going on.
    for OneParm in OneActor.Parameters():
        if OneParm.IsMorphTarget():
            OneParm.SetValue(0.0)
        scene.DrawAll()  # Show the 0 setting.

    if Name == "":
        Name = "GTM" # If you forgot a name, use this. Poser will add _1, etc auto.

    OneActor.SpawnTarget(Name)  # Now create the new
    NewParm = OneActor.Parameter(Name) # Fetch the new parm (Spawn doesn't return anything)

    # Now use the Sorted and Delta dicts to generate a new MT, only in the join or intersect zones.
    if Symm:
        for V in SortedVerts.keys():
            (dummyV,G,D,OldX, OldY, OldZ) =  SortedVerts[V]
            (dummyV,G,D,DeltaX, DeltaY, DeltaZ) =  DeltaVerts[V]
            # For each SortedVert:
            # Try to find one with (-X,Y,Z) nearly the same as the (X,Y,Z) before
            # the delta on the "working" side.  The SymmVertIndex array just carries
            # the V index for the symm vert.  In other words,
            # SymmVertIndex[MainVert] = Index of symm vert
            for S in range(OneGeom.NumVertices()):
                if S not in DeltaVerts.keys(): # don't look at main verts
                    if S not in SymmVertIndexes.keys(): # don't repeat
                        SymVert = OneGeom.Vertex(S) # Get the vert obj
                        SX = SymVert.X()
                        SY = SymVert.Y()
                        SZ = SymVert.Z()
                        # Try to locate the nearest vert to (-OldX,OldY,OldZ)
                        if abs((-SX)-OldX) < CoarseInc:
                            if abs(  SY-OldY) < CoarseInc:
                                if abs(  SZ-OldZ) < CoarseInc:
                                    SymmVertIndexes[V] = S  # Hold for smoothing
                                    NewParm.SetMorphTargetDelta(S, -DeltaX, DeltaY, DeltaZ)
    else:  # Regular, not symm
        for V in SortedVerts.keys():
            (dummyV,G,D,OldX, OldY, OldZ) =  SortedVerts[V]
            (dummyV,G,D,DeltaX, DeltaY, DeltaZ) =  DeltaVerts[V]
            NewParm.SetMorphTargetDelta(V, DeltaX, DeltaY, DeltaZ)
            # Put the originals back so that the MT can honestly start from 0.
            OneVert = OneGeom.Vertex(V)
            OneVert.SetX(OldX)
            OneVert.SetY(OldY)
            OneVert.SetZ(OldZ)

#------------------------------------------------------------

def DiffDist(Base1,Base2,Point):
    # each arg is a 3-tuple
    B1X,B1Y,B1Z = Base1
    B2X,B2Y,B2Z = Base2
    X,Y,Z = Point

    # First get the dist between Base1 and Base2.
    DX=B2X-B1X
    DY=B2Y-B1Y
    DZ=B2Z-B1Z
    Dist12 = math.sqrt(DX*DX+DY*DY+DZ*DZ)

    # Dist of Point from Midpoint:
    DX = X - (B1X+B2X)/2.0
    DY = Y - (B1Y+B2Y)/2.0
    DZ = Z - (B1Z+B2Z)/2.0
    DistPM = math.sqrt(DX*DX+DY*DY+DZ*DZ)

    PropDist  = DistPM / Dist12
    return PropDist

#------------------------------------------------------------

def MarkLayers():
    # New approach.  Get a bounding box for the group;
    # determine distance of each vert from each corner of the box.
    # Points that are closest to center (least difference between the
    # corner distances) get the largest delta.
    global SortedVerts

    for G in SelectedGroups.keys(): # Group index
        VertsInGroup={}  # key is vert number, value is net distance.
        AvDists=[] # so we can sort. Values are tuples: AvDist,Vert number.
        # First action in each group: do a max-min on all three dims to locate corners.
        MaxX=-10000.0
        MaxY=-10000.0
        MaxZ=-10000.0
        MinX= 10000.0
        MinY= 10000.0
        MinZ= 10000.0
        VertsInGroup={}  # refresh
        for (iV,iG,iD,iX,iY,iZ) in SortedVerts.values():
            if iG == G:  # In the group
                VertsInGroup[iV]=(iV,iG,iD,iX,iY,iZ)
                if   iX < MinX: MinX=iX
                elif iX > MaxX: MaxX=iX
                if   iY < MinY: MinY=iY
                elif iY > MaxY: MaxY=iY
                if   iZ < MinZ: MinZ=iZ
                elif iZ > MaxZ: MaxZ=iZ
        # Get the 8 corners:
        FrontNW = (MinX,MaxY,MaxZ)
        FrontNE = (MaxX,MaxY,MaxZ)
        FrontSE = (MaxX,MinY,MaxZ)
        FrontSW = (MinX,MinY,MaxZ)
        BackNW =  (MinX,MaxY,MinZ)
        BackNE =  (MaxX,MaxY,MinZ)
        BackSE =  (MaxX,MinY,MinZ)
        BackSW =  (MinX,MinY,MinZ)
        # Now run through the VertsInGroup and determine the
        # net distance from all corners.
        for V in VertsInGroup.keys():
            (iV,iG,iD,iX,iY,iZ) = VertsInGroup[V]

            # Find a proportion along each diagonal:
            P = (iX,iY,iZ)
            PD1 = DiffDist(FrontNW,BackSE,P)
            PD2 = DiffDist(FrontNE,BackSW,P)
            PD3 = DiffDist(FrontSE,BackNW,P)
            PD4 = DiffDist(FrontSW,BackNE,P)
            # Get the average of these:
            AvDist = (PD1+PD2+PD3+PD4)/4.0
            AvDists.append((AvDist,V)) # So we can sort in next stage

        # Normalize the AvDists to run from 0 to 1.
        AvDists.sort()
        (BotVal,BotV) = AvDists[0]  # least dist
        (TopVal,TopV) = AvDists[-1] # biggest dist
        ActualRange = TopVal - BotVal

        # Now put the reproportioned dist back into the Sorted.
        for (Val,V) in AvDists:
            (iV,iG,iD,iX,iY,iZ) = SortedVerts[V]
            SortedVerts[V] = (iV,iG,((ActualRange-Val)/ActualRange),iX,iY,iZ)

#------------------------------------------------------------
#------------------------------------------------------------
# Next section is TK
#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        self.master.title("G2M")
        # Status goes on top
        self.StatusEntry=Entry(self.master,width=40)
        self.StatusEntry.grid(row=0,column=0,columnspan=3)
        self.StatusEntry.insert(0,"Select groups then hit Straight, Ridge, or Round.")

        # Frames for list and buttons
        self.ListFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ListFrame.grid(row=1,column=0,pady=1)
        self.FalloffFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
        self.FalloffFrame.grid(row=2,column=0,pady=1)
        self.TranFrame = Frame(self.master,bg="beige",borderwidth=2,relief=RIDGE)
        self.TranFrame.grid( row=3,column=0,pady=1)
        self.ScaleFrame = Frame(self.master,bg="lightblue",borderwidth=2,relief=RIDGE)
        self.ScaleFrame.grid( row=4,column=0,pady=1)
        self.BulgeFrame = Frame(self.master,bg="pink",borderwidth=2,relief=RIDGE)
        self.BulgeFrame.grid( row=5,column=0,pady=1)
        self.BottomFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
        self.BottomFrame.grid(row=6,column=0,pady=1)

        # List and scroll
        self.GroupScroll  = Scrollbar(self.ListFrame, orient=VERTICAL)
        self.GroupScroll.grid( row=0, column=0,sticky=N+S+E)
        self.GroupBox  = Listbox(self.ListFrame, height=5, selectmode=MULTIPLE,yscrollcommand = self.GroupScroll.set)
        self.GroupBox.grid( row=0, column=1)
        self.GroupJoin = Button(self.ListFrame,  text="Straight",command=self.handleGroupJoin)
        self.GroupJoin.grid (row=1, column=0)
        self.GroupRidge = Button(self.ListFrame,  text="Ridge",  command=self.handleGroupRidge)
        self.GroupRidge.grid (row=1, column=1)
        self.GroupRound  = Button(self.ListFrame,text="Round",command=self.handleGroupRound)
        self.GroupRound.grid(row=1, column=2)

        self.Fwidth=120
        self.Fheight=40
        self.Falloff = Canvas(self.FalloffFrame,bg="white",width=self.Fwidth,height=self.Fheight)
        self.Falloff.grid(row=0, column=0,columnspan=3)
        Label(self.FalloffFrame,text="Rounding shape").grid(row=1,column=0)
        self.FalloffSlider=Scale(self.FalloffFrame, from_=0.0,to=100.0,orient=HORIZONTAL,state=DISABLED)
        self.FalloffSlider.grid(row=1,column=1)
        self.FalloffSlider.set(50.0)
        self.FalloffSlider.bind('<ButtonRelease-1>',self.MoveFalloff)

        # Fill with names of Groups
        for OneName in OneGeom.Groups():
            self.GroupBox.insert(END,OneName)
        # Set the scrollbar to work
        self.GroupScroll["command"] = self.GroupBox.yview

        # Coarse and fine buttons
        self.XtranLeftCoarse  = Button(self.TranFrame,text="<<",command=self.handleXtranLeftCoarse,state=DISABLED)
        self.XtranLeftFine    = Button(self.TranFrame,text="<" ,command=self.handleXtranLeftFine,state=DISABLED)
        self.XtranLabel       = Label (self.TranFrame,text="X tran",bg="beige")
        self.XtranRightFine   = Button(self.TranFrame,text=">" ,command=self.handleXtranRightFine,state=DISABLED)
        self.XtranRightCoarse = Button(self.TranFrame,text=">>",command=self.handleXtranRightCoarse,state=DISABLED)
        self.XtranLeftCoarse.grid (row=0, column=0)
        self.XtranLeftFine.grid   (row=0, column=1)
        self.XtranLabel.grid      (row=0, column=2)
        self.XtranRightFine.grid  (row=0, column=3)
        self.XtranRightCoarse.grid(row=0, column=4)

        self.YtranLeftCoarse  = Button(self.TranFrame,text="<<",command=self.handleYtranLeftCoarse,state=DISABLED)
        self.YtranLeftFine    = Button(self.TranFrame,text="<" ,command=self.handleYtranLeftFine,state=DISABLED)
        self.YtranLabel       = Label (self.TranFrame,text="Y tran",bg="beige")
        self.YtranRightFine   = Button(self.TranFrame,text=">" ,command=self.handleYtranRightFine,state=DISABLED)
        self.YtranRightCoarse = Button(self.TranFrame,text=">>",command=self.handleYtranRightCoarse,state=DISABLED)
        self.YtranLeftCoarse.grid (row=1, column=0)
        self.YtranLeftFine.grid   (row=1, column=1)
        self.YtranLabel.grid      (row=1, column=2)
        self.YtranRightFine.grid  (row=1, column=3)
        self.YtranRightCoarse.grid(row=1, column=4)

        self.ZtranLeftCoarse  = Button(self.TranFrame,text="<<",command=self.handleZtranLeftCoarse,state=DISABLED)
        self.ZtranLeftFine    = Button(self.TranFrame,text="<" ,command=self.handleZtranLeftFine,state=DISABLED)
        self.ZtranLabel       = Label (self.TranFrame,text="Z tran",bg="beige")
        self.ZtranRightFine   = Button(self.TranFrame,text=">" ,command=self.handleZtranRightFine,state=DISABLED)
        self.ZtranRightCoarse = Button(self.TranFrame,text=">>",command=self.handleZtranRightCoarse,state=DISABLED)
        self.ZtranLeftCoarse.grid (row=2, column=0)
        self.ZtranLeftFine.grid   (row=2, column=1)
        self.ZtranLabel.grid      (row=2, column=2)
        self.ZtranRightFine.grid  (row=2, column=3)
        self.ZtranRightCoarse.grid(row=2, column=4)


        self.XscaleLeftCoarse  = Button(self.ScaleFrame,text="<<",command=self.handleXscaleLeftCoarse,state=DISABLED)
        self.XscaleLeftFine    = Button(self.ScaleFrame,text="<" ,command=self.handleXscaleLeftFine,state=DISABLED)
        self.XscaleLabel       = Label (self.ScaleFrame,text="X scale",bg="lightblue")
        self.XscaleRightFine   = Button(self.ScaleFrame,text=">" ,command=self.handleXscaleRightFine,state=DISABLED)
        self.XscaleRightCoarse = Button(self.ScaleFrame,text=">>",command=self.handleXscaleRightCoarse,state=DISABLED)
        self.XscaleLeftCoarse.grid (row=0, column=0)
        self.XscaleLeftFine.grid   (row=0, column=1)
        self.XscaleLabel.grid      (row=0, column=2)
        self.XscaleRightFine.grid  (row=0, column=3)
        self.XscaleRightCoarse.grid(row=0, column=4)

        self.YscaleLeftCoarse  = Button(self.ScaleFrame,text="<<",command=self.handleYscaleLeftCoarse,state=DISABLED)
        self.YscaleLeftFine    = Button(self.ScaleFrame,text="<" ,command=self.handleYscaleLeftFine,state=DISABLED)
        self.YscaleLabel       = Label (self.ScaleFrame,text="Y scale",bg="lightblue")
        self.YscaleRightFine   = Button(self.ScaleFrame,text=">" ,command=self.handleYscaleRightFine,state=DISABLED)
        self.YscaleRightCoarse = Button(self.ScaleFrame,text=">>",command=self.handleYscaleRightCoarse,state=DISABLED)
        self.YscaleLeftCoarse.grid (row=1, column=0)
        self.YscaleLeftFine.grid   (row=1, column=1)
        self.YscaleLabel.grid      (row=1, column=2)
        self.YscaleRightFine.grid  (row=1, column=3)
        self.YscaleRightCoarse.grid(row=1, column=4)

        self.ZscaleLeftCoarse  = Button(self.ScaleFrame,text="<<",command=self.handleZscaleLeftCoarse,state=DISABLED)
        self.ZscaleLeftFine    = Button(self.ScaleFrame,text="<" ,command=self.handleZscaleLeftFine,state=DISABLED)
        self.ZscaleLabel       = Label (self.ScaleFrame,text="Z scale",bg="lightblue")
        self.ZscaleRightFine   = Button(self.ScaleFrame,text=">" ,command=self.handleZscaleRightFine,state=DISABLED)
        self.ZscaleRightCoarse = Button(self.ScaleFrame,text=">>",command=self.handleZscaleRightCoarse,state=DISABLED)
        self.ZscaleLeftCoarse.grid (row=2, column=0)
        self.ZscaleLeftFine.grid   (row=2, column=1)
        self.ZscaleLabel.grid      (row=2, column=2)
        self.ZscaleRightFine.grid  (row=2, column=3)
        self.ZscaleRightCoarse.grid(row=2, column=4)

        self.BulgeLeftCoarse  = Button(self.BulgeFrame,text="<<",command=self.handleBulgeLeftCoarse,state=DISABLED)
        self.BulgeLeftFine    = Button(self.BulgeFrame,text="<" ,command=self.handleBulgeLeftFine,state=DISABLED)
        self.BulgeLabel       = Label (self.BulgeFrame,text=" Bulge",bg="pink")
        self.BulgeRightFine   = Button(self.BulgeFrame,text=">" ,command=self.handleBulgeRightFine,state=DISABLED)
        self.BulgeRightCoarse = Button(self.BulgeFrame,text=">>",command=self.handleBulgeRightCoarse,state=DISABLED)
        self.BulgeLeftCoarse.grid (row=0, column=0)
        self.BulgeLeftFine.grid   (row=0, column=1)
        self.BulgeLabel.grid      (row=0, column=2)
        self.BulgeRightFine.grid  (row=0, column=3)
        self.BulgeRightCoarse.grid(row=0, column=4)

        # Name and Go buttons, in Bottom frame
        self.NameLabel = Label(self.BottomFrame,text="Name for new MT:")
        self.NameLabel.grid(row=0, column=0)
        self.NameEntry = Entry(self.BottomFrame,state=DISABLED)
        self.NameEntry.grid(row=0, column=1)
        self.buttonNormGo = Button(self.BottomFrame, text="Go", command=self.handleNormGo,state=DISABLED)
        self.buttonNormGo.grid(row=1, column=0)
        self.buttonSymmGo = Button(self.BottomFrame, text="Mirror", command=self.handleSymmGo,state=DISABLED)
        self.buttonSymmGo.grid(row=1, column=1)
        self.buttonCancel = Button(self.BottomFrame, text="Cancel", command=self.handleCancel)
        self.buttonCancel.grid(row=1, column=2)

        self.master.protocol('WM_DELETE_WINDOW', self.handleCancel)
        self.MoveFalloff(None)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def MoveFalloff(self,event):
        Val = self.FalloffSlider.get() * 0.004
        X1=3
        X2=self.Fwidth-3
        Y1=3
        Y2=self.Fheight-3
        # define left end of mesa
        LX = Val * self.Fwidth
        LY = Y2/2 # Put first critpoint at bottom
        # Center point, static.
        CX = X1 + (X2+X1)/2
        CY = Y1 # High
        # define right end of mesa
        RX = self.Fwidth-(Val * self.Fwidth)
        RY = Y2/2 # Put second critpoint at bottom
        # Draw line from SW corner (X1,Y2) to centerpoint (CX,CY) to NE corner (X2,Y1)
        self.Falloff.delete("all")
        self.Falloff.create_line(X1,Y2, LX,LY, CX,CY, RX,RY, X2,Y2,width=3,smooth=1)

    # - - - - - - - - - - - - - - - - - -

    def Round(self,NewName,Symm):
        # Go through the verts in the already-made MT. For each, interpolate by
        # the distances found in MarkLayers.
        global DeltaVerts
        global SymmVertIndexes
        try:
            NewParm = OneActor.Parameter(NewName) # Re-fetch the MT parm
        except:
            raise "Given name for MT doesn't work."
        # Figure proportion from the D factor.
        # Include the falloff and tanh.  The D here was formed in the routine above,
        # so that it is 1 at center and 0 nearest edge.  To give this a movable
        # ogive, we need to move the domain of the tanh.  When the slider is at max,
        # which has the sharpest center, we move the domain so that 0 on the tanh arg
        # hits at about 0.95 on D; when the slider is at min, 0 on tanh arg hits
        # at about 0.05 on D.  So, Result = 0.5 * (1.0+math.tanh(D*2-Offset))
        Offset = self.FalloffSlider.get() * 0.01

        for V in SortedVerts.keys():
            if Symm:
                try:
                   rV,G,D,X,Y,Z = SortedVerts[V] # We only need the dist here
                   S = SymmVertIndexes[V]
                   DeltaX,DeltaY,DeltaZ = NewParm.MorphTargetDelta(S)
                   NetD = 100.0 * (1.0 + math.tanh( D * 2.0 - Offset))
                   NewParm.SetMorphTargetDelta(S, DeltaX*NetD, DeltaY*NetD, DeltaZ*NetD)
                except:
                   pass
            else:
                rV,G,D,X,Y,Z = SortedVerts[V] # We only need the dist here
                DeltaX,DeltaY,DeltaZ = NewParm.MorphTargetDelta(V)
                NetD = 100.0 * (1.0 + math.tanh( D * 2.0 - Offset))
                # This gives a good shape, but the edge moves too far.  Need to
                # slide it over so that the edge verts don't move at all.
                try:
                    NewParm.SetMorphTargetDelta(V, DeltaX*NetD, DeltaY*NetD, DeltaZ*NetD)
                except:
                    pass

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        scene.ProcessSomeEvents(1)
        root.lift()
        root.after(100, self.Update)

    # - - - - - - - - - - - - - - - - - -
    def handleXtranLeftCoarse(self):
        DisplaceX(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXtranLeftFine(self):
        DisplaceX(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXtranRightFine(self):
        DisplaceX(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXtranRightCoarse(self):
        DisplaceX(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleYtranLeftCoarse(self):
        DisplaceY(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYtranLeftFine(self):
        DisplaceY(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYtranRightFine(self):
        DisplaceY(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYtranRightCoarse(self):
        DisplaceY(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleZtranLeftCoarse(self):
        DisplaceZ(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZtranLeftFine(self):
        DisplaceZ(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZtranRightFine(self):
        DisplaceZ(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZtranRightCoarse(self):
        DisplaceZ(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleBulgeLeftCoarse(self):
        DisplaceBulge(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleBulgeLeftFine(self):
        DisplaceBulge(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleBulgeRightFine(self):
        DisplaceBulge(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleBulgeRightCoarse(self):
        DisplaceBulge(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleXscaleLeftCoarse(self):
        DisplaceXscale(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXscaleLeftFine(self):
        DisplaceXscale(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXscaleRightFine(self):
        DisplaceXscale(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleXscaleRightCoarse(self):
        DisplaceXscale(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleYscaleLeftCoarse(self):
        DisplaceYscale(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYscaleLeftFine(self):
        DisplaceYscale(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYscaleRightFine(self):
        DisplaceYscale(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleYscaleRightCoarse(self):
        DisplaceYscale(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -
    def handleZscaleLeftCoarse(self):
        DisplaceZscale(-CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZscaleLeftFine(self):
        DisplaceZscale(-FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZscaleRightFine(self):
        DisplaceZscale(+FineInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -
    def handleZscaleRightCoarse(self):
        DisplaceZscale(+CoarseInc)
        scene.DrawAll()
    # - - - - - - - - - - - - - - - - - -

    def Unhide(self):
        # Unhide the major buttons
        self.ShowStatus("Use the buttons below to move groups.")

        self.FalloffSlider.configure(state=NORMAL)

        self.XtranLeftCoarse.configure(state=NORMAL)
        self.XtranLeftFine.configure(state=NORMAL)
        self.XtranRightFine.configure(state=NORMAL)
        self.XtranRightCoarse.configure(state=NORMAL)

        self.YtranLeftCoarse.configure(state=NORMAL)
        self.YtranLeftFine.configure(state=NORMAL)
        self.YtranRightFine.configure(state=NORMAL)
        self.YtranRightCoarse.configure(state=NORMAL)

        self.ZtranLeftCoarse.configure(state=NORMAL)
        self.ZtranLeftFine.configure(state=NORMAL)
        self.ZtranRightFine.configure(state=NORMAL)
        self.ZtranRightCoarse.configure(state=NORMAL)

        self.BulgeLeftCoarse.configure(state=NORMAL)
        self.BulgeLeftFine.configure(state=NORMAL)
        self.BulgeRightFine.configure(state=NORMAL)
        self.BulgeRightCoarse.configure(state=NORMAL)

        self.XscaleLeftCoarse.configure(state=NORMAL)
        self.XscaleLeftFine.configure(state=NORMAL)
        self.XscaleRightFine.configure(state=NORMAL)
        self.XscaleRightCoarse.configure(state=NORMAL)

        self.YscaleLeftCoarse.configure(state=NORMAL)
        self.YscaleLeftFine.configure(state=NORMAL)
        self.YscaleRightFine.configure(state=NORMAL)
        self.YscaleRightCoarse.configure(state=NORMAL)

        self.ZscaleLeftCoarse.configure(state=NORMAL)
        self.ZscaleLeftFine.configure(state=NORMAL)
        self.ZscaleRightFine.configure(state=NORMAL)
        self.ZscaleRightCoarse.configure(state=NORMAL)

        self.NameEntry.configure(state=NORMAL)
        self.buttonNormGo.configure(state=NORMAL)
        self.buttonSymmGo.configure(state=NORMAL)


    # - - - - - - - - - - - - - - - - - -

    def handleGroupJoin(self):
        # Pick up the group names and make the regular buttons visible.
        global SelectedGroups
        global Smoothable
        try:
            SelectedIndexes = self.GroupBox.curselection()
            for N in SelectedIndexes:
                intN = int(N)
                SelectedGroups[intN] = self.GroupBox.get(intN)
        except:
            pass  # we already set default to GroupList[0]
        self.Unhide()
        # Set up the real action.
        LocateAllVerts()
        SortForJoin()
        Smoothable = 0 # Force no smooth

    # - - - - - - - - - - - - - - - - - -

    def handleGroupRidge(self):
        # Pick up the group names and make the regular buttons visible.
        global SelectedGroups
        global Smoothable
        try:
            SelectedIndexes = self.GroupBox.curselection()
            for N in SelectedIndexes:
                intN = int(N)
                SelectedGroups[intN] = self.GroupBox.get(intN)
        except:
            pass  # we already set default to GroupList[0]
        self.Unhide()
        # Set up the real action.
        LocateAllVerts()
        SortForRidge()
        Smoothable = 0 # Force no smooth

    # - - - - - - - - - - - - - - - - - -

    def handleGroupRound(self):
        # Pick up the group names and make the regular buttons visible.
        global SelectedGroups
        global Smoothable
        try:
            SelectedIndexes = self.GroupBox.curselection()
            for N in SelectedIndexes:
                intN = int(N)
                SelectedGroups[intN] = self.GroupBox.get(intN)
        except:
            pass  # we already set default to GroupList[0]
        self.Unhide()
        # Set up the real action.
        LocateAllVerts()
        SortForJoin()
        Smoothable = 1 # Force smooth

    # - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - -

    def handleNormGo(self):
        Symm = 0
        # Get the name:
        NewName = self.NameEntry.get()
        if not NewName: NewName = 'G2M'
        SpawnResult(NewName,Symm)
        if Smoothable:
           Count = MarkLayers()
           self.Round(NewName,Symm)
        self.master.destroy()
        self.master.quit()

    # - - - - - - - - - - - - - - - - - -

    def handleSymmGo(self):
        Symm = 1
        # Get the name:
        NewName = self.NameEntry.get()
        if not NewName: NewName = 'G2M'
        SpawnResult(NewName,Symm)
        if Smoothable:
            MarkLayers()
            self.Round(NewName,Symm)
        self.master.destroy()
        self.master.quit()

    # - - - - - - - - - - - - - - - - - -

    def handleCancel(self):
        self.master.destroy()
        self.master.quit()

#------------------------------------------------------------
# End defs, begin action here.
#------------------------------------------------------------

InitFirst()
root = Tk()
app = App(root,'')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
